@@ -12,8 +12,9 @@ use Sys::Hostname 'hostname';
 use Time::Piece ();
 
 has 'mysql';
+has 'no_txn' => sub { 0 };
 
-our $VERSION = '0.25';
+our $VERSION = '0.26';
 
 sub dequeue {
   my ($self, $worker_id, $wait, $options) = @_;
@@ -78,33 +79,55 @@ sub enqueue {
   my $args    = shift // [];
   my $options = shift // {};
 
+  # Pre-compute parameters to reduce time holding DB transaction
+  my @insert_params = (
+    encode_json( $args ),
+    $options->{attempts} // 1,
+    $options->{delay} // 0,
+    ($options->{expire})x2,
+    $options->{lax} ? 1 : 0,
+    $options->{priority} // 0,
+    $options->{queue} // 'default',
+    $task,
+  );
+
+  my @parents = @{ $options->{parents} // [] };
+  my %notes = %{ $options->{notes} // {} };
+
+  my ( $insert_parents_sql, $insert_notes_sql );
+  if ( @parents ) {
+    $insert_parents_sql = "INSERT IGNORE INTO minion_jobs_depends (`parent_id`, `child_id`) VALUES "
+      . join( ", ", map "( ?, ? )", @parents );
+  }
+  if ( keys %notes ) {
+    $insert_notes_sql = 'INSERT INTO minion_notes (`job_id`, `note_key`, `note_value`) VALUES '
+      . join( ', ', map '( ?, ?, ? )', keys %notes );
+    $notes{ $_ } = encode_json( $notes{ $_ } ) for keys %notes;
+  }
+
   my $db = $self->mysql->db;
   # Use a transaction to commit the entire job at once. Without this,
   # a job may be started before one of its parents, since the parent
   # restriction is added after the job itself.
-  my $tx = $db->begin;
+  my $tx = $self->no_txn ? undef : $db->begin;
 
   my $job_id = $db->query(
     "insert into minion_jobs (`args`, `attempts`, `delayed`, `expires`, `lax`, `priority`, `queue`, `task`)
      values (?, ?, (DATE_ADD(NOW(), INTERVAL ? SECOND)), case when ? is not null then date_add( now(), interval ? second ) end, ?, ?, ?, ?)",
-     encode_json($args), $options->{attempts} // 1,
-     $options->{delay} // 0, ($options->{expire})x2, $options->{lax} ? 1 : 0,
-     $options->{priority} // 0, $options->{queue} // 'default', $task,
+     @insert_params,
   )->last_insert_id;
 
-if ( my $notes = $options->{notes} ) {
-    $self->_note( $job_id, $notes, $db );
+  if ( $insert_parents_sql ) {
+    my @insert_parents_params = map { $_, $job_id  } @parents;
+    $db->query( $insert_parents_sql, @insert_parents_params );
   }
 
-  if ( my @parents = @{ $options->{parents} || [] } ) {
-    $db->query(
-      "INSERT IGNORE INTO minion_jobs_depends (`parent_id`, `child_id`) VALUES "
-      . join( ", ", map "( ?, ? )", @parents ),
-      map { $_, $job_id  } @parents
-    );
+  if ( $insert_notes_sql ) {
+    my @insert_notes_params = map { $job_id, $_, $notes{$_} } keys %notes;
+    $db->query( $insert_notes_sql, @insert_notes_params );
   }
 
-  $tx->commit;
+  $tx->commit if defined $tx;
   $self->mysql->pubsub->notify("minion.job" => $job_id);
 
   return $job_id;
@@ -112,28 +135,30 @@ if ( my $notes = $options->{notes} ) {
 
 sub note {
   my ($self, $id, $notes) = @_;
-  return $self->_note( $id, $notes, $self->mysql->db );
-}
-
-sub _note {
-  my ($self, $id, $notes, $db) = @_;
 
   my @replace_keys = grep defined $notes->{ $_ }, keys %$notes;
   my @delete_keys = grep !defined $notes->{ $_ }, keys %$notes;
 
-  my $replaced = !!eval {
-    $db->query(
-      'REPLACE INTO minion_notes (`job_id`, `note_key`, `note_value`) VALUES '
-      . join( ', ', map '( ?, ?, ? )', @replace_keys ),
-      map { $id, $_, encode_json( $notes->{$_} ) } @replace_keys
+  my ( $replaced, $deleted );
+  my $db = $self->mysql->db;
+  if ( @replace_keys ) {
+    local $@; # Don't clobber global
+    $replaced = !!eval {
+      $db->query(
+        'REPLACE INTO minion_notes (`job_id`, `note_key`, `note_value`) VALUES '
+        . join( ', ', map '( ?, ?, ? )', @replace_keys ),
+        map { $id, $_, encode_json( $notes->{$_} ) } @replace_keys
+      )->rows;
+    };
+  }
+  if ( @delete_keys ) {
+    $deleted = !!$db->delete(
+      minion_notes => {
+        job_id => $id,
+        note_key => { -in => \@delete_keys },
+      }
     )->rows;
-  };
-  my $deleted = !!$db->delete(
-    minion_notes => {
-      job_id => $id,
-      note_key => { -in => \@delete_keys },
-    }
-  )->rows;
+  }
 
   return $replaced || $deleted;
 }
@@ -432,6 +457,8 @@ sub reset {
     $db->query("ALTER TABLE minion_jobs AUTO_INCREMENT = 1");
     $db->query("truncate table minion_locks");
     $db->query("ALTER TABLE minion_locks AUTO_INCREMENT = 1");
+    $db->query("truncate table minion_jobs_depends");
+    $db->query("truncate table minion_notes");
     $db->query("truncate table minion_workers");
     $db->query("ALTER TABLE minion_workers AUTO_INCREMENT = 1");
   }
@@ -548,52 +575,62 @@ sub _try {
 
   my $qq = join ", ", map({ "?" } @$queues);
   my $qt = join ", ", map({ "?" } @$tasks );
+  my @bind_vars = ( @$queues, @$tasks );
+
+  my $sql_job_id = '';
+  if ( defined $options->{id} ) {
+    $sql_job_id = 'AND job.id = ?';
+    push @bind_vars, $options->{id};
+  }
+
+  my $sql = qq{
+    SELECT job.id, job.args, job.retries, job.task
+    FROM minion_jobs job
+    LEFT JOIN minion_jobs_depends depends ON depends.child_id = job.id
+    LEFT JOIN minion_jobs parent ON parent.id = depends.parent_id
+    WHERE job.state = 'inactive'
+      AND job.`delayed` <= NOW()
+      AND job.queue IN ($qq) AND job.task IN ($qt)
+      $sql_job_id
+      AND (job.expires IS NULL OR job.expires > NOW())
+    GROUP BY job.id
+    HAVING SUM(
+      parent.state IS NOT NULL AND (
+        parent.state = 'active'
+        OR ( parent.state = 'failed' AND NOT job.lax )
+        OR ( parent.state = 'inactive' AND (parent.expires IS NULL OR parent.expires > NOW()))
+      )
+    ) = 0
+    ORDER BY job.priority DESC, job.created
+    LIMIT 1
+  };
 
   my $db = $self->mysql->db;
-  my $tx = $db->begin;
+  my $i = 5; # Try this many times to get a job
+  my $job;
+  while ( $i-- > 0 ) {
+    # Find a candidate job to run
+    my $res = $db->query( $sql, @bind_vars )->hash;
+    return unless $res; # No jobs ready to work on
 
-  my $res = $db->query(
-    qq{
-      SELECT job.id, job.args, job.retries, job.task
-      FROM minion_jobs job
-      WHERE job.state = 'inactive'
-        AND job.`delayed` <= NOW()
-        AND NOT EXISTS (
-          SELECT 1 FROM minion_jobs_depends depends
-          LEFT JOIN (
-            SELECT id, state, expires
-            FROM minion_jobs
-          ) AS parent ON parent.id=depends.parent_id
-          WHERE child_id=job.id
-            AND (
-              parent.state = 'active'
-              OR ( parent.state = 'failed' AND NOT job.lax )
-              OR ( parent.state = 'inactive' AND (parent.expires IS NULL OR parent.expires > NOW()))
-          )
-        )
-        AND job.id = COALESCE(?, job.id) AND job.queue IN ($qq) AND job.task IN ($qt)
-        AND (expires IS NULL OR expires > NOW())
-      ORDER BY job.priority DESC, job.created
-      LIMIT 1
-      FOR UPDATE
-    },
-    $options->{id}, @$queues, @$tasks,
-  );
+    # Try to claim the job for this worker. There is a race condition
+    # between selecting the job and claiming it, so we need to make sure
+    # we were the one to claim it.
+    my $claimed = $db->query(
+      qq{ UPDATE minion_jobs SET started = NOW(), state = 'active', worker = ? WHERE id = ? AND state = 'inactive' },
+      $worker_id, $res->{id},
+    )->affected_rows;
 
-  my $job = $res->hash;
-  return if !$job || !%$job;
+    # Try again if we lose the race
+    next if !$claimed;
 
-  $db->dbh->do(
-    qq{
-      UPDATE minion_jobs SET started = NOW(), state = 'active', worker = ?
-      WHERE id = ?
-    },
-    {}, $worker_id, $job->{id},
-  );
-  $tx->commit;
+    # Won the race, so do the job
+    $job = $res;
+    last;
+  }
 
   #; use Data::Dumper;
-  #; say "Dequeuing job: " . Dumper $job;
+  #; say "Dequeued job: " . Dumper $job;
 
   $job->{args} = $job->{args} ? decode_json($job->{args}) : undef;
 
@@ -732,6 +769,16 @@ sub _migrate_notes {
 #pod
 #pod L<Mojo::mysql> object used to store all data.
 #pod
+#pod =head2 no_txn
+#pod
+#pod If true, will not make a transaction around the L</enqueue> queries.
+#pod Without a transaction, a job could be dequeued before its parent
+#pod relationships are written to the database. However, since MySQL does not
+#pod support nested transactions (despite supporting something almost exactly
+#pod like them...), you can disable transactions for testing by setting this
+#pod attribute (if you perform your tests in a transaction so they can be
+#pod rolled back when the test is complete).
+#pod
 #pod =head1 METHODS
 #pod
 #pod L<Minion::Backend::mysql> inherits all methods from L<Minion::Backend> and
@@ -1104,7 +1151,7 @@ Minion::Backend::mysql
 
 =head1 VERSION
 
-version 0.25
+version 0.26
 
 =head1 SYNOPSIS
 
@@ -1150,6 +1197,16 @@ implements the following new ones.
 
 L<Mojo::mysql> object used to store all data.
 
+=head2 no_txn
+
+If true, will not make a transaction around the L</enqueue> queries.
+Without a transaction, a job could be dequeued before its parent
+relationships are written to the database. However, since MySQL does not
+support nested transactions (despite supporting something almost exactly
+like them...), you can disable transactions for testing by setting this
+attribute (if you perform your tests in a transaction so they can be
+rolled back when the test is complete).
+
 =head1 METHODS
 
 L<Minion::Backend::mysql> inherits all methods from L<Minion::Backend> and
@@ -1570,7 +1627,7 @@ Zoffix Znet <cpan@zoffix.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2017 by Doug Bell and Brian Medley.
+This software is copyright (c) 2021 by Doug Bell and Brian Medley.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
@@ -1739,3 +1796,10 @@ ALTER TABLE minion_notes MODIFY note_value MEDIUMBLOB;
 DROP INDEX minion_jobs_state_idx ON minion_jobs;
 CREATE INDEX minion_jobs_state_idx ON minion_jobs (state);
 ALTER TABLE minion_notes MODIFY note_value TEXT;
+
+-- 10 up
+CREATE INDEX minion_notes_note_key ON minion_notes (note_key);
+
+-- 10 down
+DROP INDEX minion_notes_note_key ON minion_notes;
+
